"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _asyncToGenerator2 = require("babel-runtime/helpers/asyncToGenerator");

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _keys = require("babel-runtime/core-js/object/keys");

var _keys2 = _interopRequireDefault(_keys);

var _bluebird = require("bluebird");

var _bluebird2 = _interopRequireDefault(_bluebird);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*  the API mixin class  */
class gtsEvaluation {
    /*  parse the operation string  */
    __opParse(op) {
        let m = op.match(/^(read|create|update|delete):(direct|relation):(one|many|all)$/);
        if (m === null) throw new Error(`internal error: invalid operation "${op}"`);
        return { action: m[1], via: m[2], onto: m[3] };
    }

    /*  check whether scope has a particular operation  */
    __recordsContainOp(records, cb) {
        let types = (0, _keys2.default)(records);
        for (var i = 0; i < types.length; i++) {
            let ops = (0, _keys2.default)(records[types[i]]);
            for (var j = 0; j < ops.length; j++) {
                let op = this.__opParse(ops[j]);
                if (cb(op)) return true;
            }
        }
        return false;
    }

    /*  does a new scope outdate an old scope  */
    __scopeOutdated(recordsNew, recordsOld) {

        /*  ==== CASE 1 ====
             "Is an old/previously read OID now (directly) written onto?"  */

        /*  find all OIDs in new scope records which write  */
        let recordsNewWriteOID = {};
        (0, _keys2.default)(recordsNew).forEach(type => {
            (0, _keys2.default)(recordsNew[type]).forEach(op => {
                let opDetail = this.__opParse(op);
                if (opDetail.action.match(/^(?:update|delete)$/)) {
                    recordsNew[type][op].forEach(oid => {
                        recordsNewWriteOID[oid] = op;
                    });
                }
            });
        });

        /*  for each old scope records which read, ...  */
        let recordsOldTypes = (0, _keys2.default)(recordsOld);
        for (let i = 0; i < recordsOldTypes.length; i++) {
            let recordsOldType = recordsOldTypes[i];
            let recordsOldOps = (0, _keys2.default)(recordsOld[recordsOldType]);
            for (let j = 0; j < recordsOldOps.length; j++) {
                let recordsOldOp = recordsOldOps[j];
                let recordsOldOpDetail = this.__opParse(recordsOldOp);
                if (recordsOldOpDetail.action === "read") {
                    /*  ...check if any of its OIDs match the write OIDs in the new scope  */
                    let recordsOldOIDs = recordsOld[recordsOldType][recordsOldOp];
                    for (let k = 0; k < recordsOldOIDs.length; k++) {
                        let oid = recordsOldOIDs[k];
                        if (recordsNewWriteOID[oid]) {
                            this.emit("scope-outdated-direct", {
                                type: recordsOldType,
                                opWrite: recordsNewWriteOID[oid],
                                opRead: recordsOldOp,
                                oid: oid
                            });
                            this.emit("debug", "scope-outdated-direct " + `type=${recordsOldType} ` + `opWrite=${recordsNewWriteOID[oid]} ` + `opRead=${recordsOldOp} ` + `oid=${oid}`);
                            return true;
                        }
                    }
                }
            }
        }

        /*  ==== CASE 2 ====
             "Has an old/previously read list of OID potentially had to take
            write results into account?". For this we have to know that the
            valid operation combinations are which can occur in practice:
                         ACTION  VIA             ONTO
                        ------- --------------- ------------
            old Scope:  read    direct|relation one|many|all
            new Scope:  create  direct          one
            new Scope:  update  direct          one|many|all
            new Scope:  delete  direct          one|many|all
             So, we have to take into account:
            (create|update|delete):*:* --outdates--> read:*:(many|all)  */

        /*  for each new scope records which write, ...  */
        let recordsNewTypes = (0, _keys2.default)(recordsNew);
        for (let i = 0; i < recordsNewTypes.length; i++) {
            let recordsNewType = recordsNewTypes[i];
            let recordsNewOps = (0, _keys2.default)(recordsNew[recordsNewType]);
            for (let j = 0; j < recordsNewOps.length; j++) {
                let recordsNewOp = recordsNewOps[j];
                let recordsNewOpDetail = this.__opParse(recordsNewOp);
                if (recordsNewOpDetail.action.match(/^(?:create|update|delete)$/)) {

                    /*  for each old scope records which read, ...  */
                    if (recordsOld[recordsNewType] === undefined) continue;
                    let recordsOldOps = (0, _keys2.default)(recordsOld[recordsNewType]);
                    for (let l = 0; l < recordsOldOps.length; l++) {
                        let recordsOldOp = recordsOldOps[l];
                        let recordsOldOpDetail = this.__opParse(recordsOldOp);
                        if (recordsOldOpDetail.action === "read") {

                            /*  check for outdate situation  */
                            if (recordsOldOpDetail.onto === "many" || recordsOldOpDetail.onto === "all") {
                                this.emit("scope-outdated-indirect", {
                                    type: recordsNewType,
                                    opWrite: recordsNewOp,
                                    opRead: recordsOldOp
                                });
                                this.emit("debug", "scope-outdated-indirect " + `type=${recordsNewType} ` + `opWrite=${recordsNewOp} ` + `opRead=${recordsOldOp}`);
                                return true;
                            }
                        }
                    }
                }
            }
        }

        /*  ...else the scope is still valid (not outdated)  */
        return false;
    }

    /*  process a committed scope  */
    __scopeProcess(scope) {
        var _this = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  determine parameters  */
            let sid = scope.sid;
            let cid = scope.connection !== null ? scope.connection.cid : `${_this.uuid}:none`;

            /*  determine whether any write operations exist in the scope records  */
            let hasWriteOps = _this.__recordsContainOp(scope.records, function (op) {
                return op.action.match(/^(?:create|update|delete)$/);
            });

            /*  queries (scopes without writes)...  */
            if (!hasWriteOps) {
                if (scope.state === "subscribed") {
                    /*  ...with subscriptions are remembered  */
                    let rec = yield _this.keyval.get(`sid:${sid},rec`);
                    if (rec === undefined) yield _this.keyval.put(`sid:${sid},rec`, scope.records);
                    let pid = yield _this.keyval.get(`sid:${sid},cid:${cid}`);
                    if (pid === undefined || pid !== process.pid) yield _this.keyval.put(`sid:${sid},cid:${cid}`, process.pid);
                    _this.emit("debug", `scope-store-update sid=${sid} cid=${cid}`);
                } else {
                    /*  ...without subscriptions can be just destroyed
                        (and the processing short-circuited)  */
                    scope.destroy();
                }
            }

            /*  mutations (scopes with writes) might outdate queries (scopes with reads)  */
            else {
                    /*  iterate over all stored scope records  */
                    let keys = yield _this.keyval.keys("sid:*,rec");
                    let sids = keys.map(function (key) {
                        return key.replace(/^sid:(.+?),rec$/, "$1");
                    });
                    let outdatedSids = [];
                    yield _bluebird2.default.each(sids, (() => {
                        var _ref = (0, _asyncToGenerator3.default)(function* (otherSid) {
                            let records = yield _this.keyval.get(`sid:${otherSid},rec`);
                            if (records !== undefined) if (_this.__scopeOutdated(scope.records, records)) outdatedSids.push(otherSid);
                        });

                        return function (_x) {
                            return _ref.apply(this, arguments);
                        };
                    })());

                    /*  externally publish ids of outdated queries to all instances
                        (comes in on all instances via __scopeOutdatedEvent below)  */
                    if (outdatedSids.length > 0) {
                        _this.emit("debug", `scope-outdated-send sids=${outdatedSids.join(",")}`);
                        _this.pubsub.publish("outdated", outdatedSids);
                    }
                }
        })();
    }

    /*  process an outdated event  */
    __scopeOutdatedEvent(sids) {
        this.emit("debug", `scope-outdated-receive sids=${sids.join(",")}`);
        this.connections.forEach(conn => {
            let outdated = {};
            conn.scopes.forEach(scope => {
                if (sids.indexOf(scope.sid) >= 0) {
                    if (scope.state === "subscribed") outdated[scope.sid] = true;else if (scope.state === "paused") scope.outdated = true;
                }
            });
            outdated = (0, _keys2.default)(outdated);
            if (outdated.length > 0) {
                this.emit("debug", `scope-outdated-notify sids=${outdated.join(",")}`);
                conn.notify(outdated);
            }
        });
    }

    /*  destroy a scope  */
    __scopeDestroy(scope) {
        var _this2 = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  determine parameters  */
            let sid = scope.sid;
            let cid = scope.connection !== null ? scope.connection.cid : `${_this2.uuid}:none`;

            /*  scope records with no more corresponding subscriptions are deleted  */
            yield _this2.keyval.del(`sid:${sid},cid:${cid}`);
            let keys = yield _this2.keyval.keys(`sid:${sid},cid:*`);
            if (keys.length === 0) yield _this2.keyval.del(`sid:${sid},rec`);
            _this2.emit("debug", `scope-store-delete sid=${sid} cid=${cid}`);
        })();
    }

    /*  dump current information  */
    dump() {
        var _this3 = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  determine information in store  */
            yield _this3.keyval.acquire();
            let info = { sids: {} };
            let keys = yield _this3.keyval.keys("sid:*,cid:*");
            keys.forEach(function (key) {
                let sid = key.replace(/^sid:(.+?),cid:.+?$/, "$1");
                let cid = key.replace(/^sid:.+?,cid:(.+?)$/, "$1");
                if (info.sids[sid] === undefined) info.sids[sid] = { cids: [], records: [] };
                info.sids[sid].cids.push(cid);
            });
            keys = yield _this3.keyval.keys("sid:*,rec");
            let sids = keys.map(function (key) {
                return key.replace(/^sid:(.+?),rec$/, "$1");
            });
            yield _bluebird2.default.each(sids, (() => {
                var _ref2 = (0, _asyncToGenerator3.default)(function* (sid) {
                    let records = yield _this3.keyval.get(`sid:${sid},rec`);
                    info.sids[sid].records.push(records);
                });

                return function (_x2) {
                    return _ref2.apply(this, arguments);
                };
            })());
            yield _this3.keyval.release();

            /*  dump information  */
            let dump = "";
            (0, _keys2.default)(info.sids).forEach(function (sid) {
                dump += `Scope { sid: ${sid} }\n`;
                info.sids[sid].cids.forEach(function (cid) {
                    dump += `    Connection { cid: ${cid} }\n`;
                });
                info.sids[sid].records.forEach(function (record) {
                    let types = (0, _keys2.default)(record);
                    for (let i = 0; i < types.length; i++) {
                        let type = types[i];
                        let ops = (0, _keys2.default)(record[type]);
                        for (let j = 0; j < ops.length; j++) {
                            let op = ops[j];
                            let oids = record[type][op];
                            for (let k = 0; k < oids.length; k++) {
                                let oid = oids[k];
                                dump += `    Record { type: ${type}, op: ${op}, oid: ${oid} }\n`;
                            }
                        }
                    }
                });
            });
            return dump;
        })();
    }

    /*  flush all information  */
    flush() {
        var _this4 = this;

        return (0, _asyncToGenerator3.default)(function* () {
            /*  flush all external storage information  */
            let keys = yield _this4.keyval.keys("sid:*,cid:*");
            yield _bluebird2.default.each(keys, (() => {
                var _ref3 = (0, _asyncToGenerator3.default)(function* (key) {
                    _this4.keyval.del(key);
                });

                return function (_x3) {
                    return _ref3.apply(this, arguments);
                };
            })());
            keys = yield _this4.keyval.keys("sid:*,rec");
            yield _bluebird2.default.each(keys, (() => {
                var _ref4 = (0, _asyncToGenerator3.default)(function* (key) {
                    _this4.keyval.del(key);
                });

                return function (_x4) {
                    return _ref4.apply(this, arguments);
                };
            })());
        })();
    }
}
exports.default = gtsEvaluation; /*
                                 **  GraphQL-Tools-Subscribe -- Subscription Framework for GraphQL-Tools
                                 **  Copyright (c) 2016-2018 Ralf S. Engelschall <rse@engelschall.com>
                                 **
                                 **  Permission is hereby granted, free of charge, to any person obtaining
                                 **  a copy of this software and associated documentation files (the
                                 **  "Software"), to deal in the Software without restriction, including
                                 **  without limitation the rights to use, copy, modify, merge, publish,
                                 **  distribute, sublicense, and/or sell copies of the Software, and to
                                 **  permit persons to whom the Software is furnished to do so, subject to
                                 **  the following conditions:
                                 **
                                 **  The above copyright notice and this permission notice shall be included
                                 **  in all copies or substantial portions of the Software.
                                 **
                                 **  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
                                 **  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
                                 **  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
                                 **  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
                                 **  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
                                 **  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
                                 **  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
                                 */

/*  external dependencies  */
